#include "MadxViewerPar.h"
/*Piotr Skowronski, CERN*/

#include "MadxViewer.h"

#include "TGraph.h"
#include "TGLabel.h"
#include "TGComboBox.h"
#include "TGNumberEntry.h"
#include "TGButton.h"
#include "TObjArray.h"
#include "TAxis.h"
#include <float.h>


ClassImp(MadxViewerPar)

const Int_t MadxViewerPar::fgkComboWidth = 120;


MadxViewerPar::MadxViewerPar(MadxViewer* master):
 fMaster(master),
 fDrawButton(0x0),
 fSameButton(0x0),
 fVarString(0x0),
 fVarFrame(0x0),
 fVarLabel(0x0),
 fVarComboBox(0x0),
 fHostFrame(0x0),
 fDrawnFunctions(0x0),
 fSaxis(0x0),
 fYaxisBuffer(0x0),
 fNSPoints(0),
 fLegend(0x0),
 fYaxis(0x0)
{
 //ctor
}

MadxViewerPar::~MadxViewerPar()
{
 //dtor
 Info("~MadxViewerPar","Deleting");
 
 fYaxis = 0x0; //to remember that it does not belong to this class

 delete  fLegend;

 if (fDrawnFunctions)
  {
    Info("~MadxViewerPar"," 1 %#x",fDrawnFunctions->At(0));
   // fDrawnFunctions->Delete();
 
    for (Int_t j = 1; j < fDrawnFunctions->GetEntriesFast(); j++)
     {
       delete fDrawnFunctions->At(j);
     }  

    Info("~MadxViewerPar"," 1-1");
    delete fDrawnFunctions;
  }

 Info("~MadxViewerPar"," 2");
 
 delete fDrawButton;
 delete fSameButton;
 delete fVarString;
 delete fVarFrame;
 delete fVarLabel;
 delete fVarComboBox;
 delete [] fSaxis;
 delete [] fYaxisBuffer;

 Info("~MadxViewerPar"," END");

}


void MadxViewerPar::Create(TGCompositeFrame* hostframe)
{

  ////////////////////////////////////////////////////
  // Variable 
  // TGHorizontalFrame
   if (hostframe == 0x0) return;
   
   fHostFrame = hostframe;
   
   fVarFrame = new TGHorizontalFrame(fHostFrame,200,30,kHorizontalFrame);
   fVarLabel = new TGLabel(fVarFrame,"Variable");
   fVarLabel->SetTextJustify(36);
   fVarFrame->AddFrame(fVarLabel, new TGLayoutHints(kLHintsLeft | kLHintsTop,10,5,2,2));
   fVarLabel->MoveResize(10,40,50,20);


   // combo box
   fVarComboBox = new TGComboBox(fVarFrame,-1,kHorizontalFrame | kSunkenFrame | kDoubleBorder | kOwnBackground);
   fVarComboBox->Resize(fgkComboWidth,20);
   fVarComboBox->Select(-1);
 
   fVarFrame->AddFrame(fVarComboBox, new TGLayoutHints(kLHintsRight | kLHintsTop,5,10,2,2));
   fVarComboBox->MoveResize(60,40,fgkComboWidth,20);
    
   fHostFrame->AddFrame(fVarFrame, new TGLayoutHints( kLHintsRight | kLHintsTop,2,2,10,5));
   fVarFrame->MoveResize(2,2,200,30);

   
  ////////////////////////////////////////////////////
  // Buttons
   

   fDrawButton = new TGTextButton(fHostFrame,"Draw");
   fDrawButton->SetTextJustify(36);
   fDrawButton->Resize(115,40);
   fHostFrame->AddFrame(fDrawButton, new TGLayoutHints(kLHintsLeft | kLHintsBottom | kLHintsExpandX ,10,10,5,5));
   fDrawButton->MoveResize(30,0,120,40);


   //Same button
   
   fSameButton = new TGCheckButton(fHostFrame,"Add to plot");
   fHostFrame->AddFrame(fSameButton, new TGLayoutHints(kLHintsLeft | kLHintsTop,2,2,2,2));
   fSameButton->MoveResize(40,288,105,24);

   fLegend =  new TLegend(0.75,0.75,0.95,0.95);
   fLegend->SetFillColor(0);

}

//______________________________________________________________________________

Bool_t MadxViewerPar::Init()
{

  Info("Init","Initializing ...");

  fNSPoints = w_ptc_getnelements(); 
  
  if (fNSPoints <= 0)
   {
     Error("Init","PTC reports no elements");
     return kFALSE;
   }

  if (fMaster == 0x0)
   {
     Error("Init","Master Viewer is not set");
     return kFALSE;
   }
  
  if (fVarComboBox == 0x0)
   {
     Error("Init","Parametric Viewer at Elements seems to be not created correctly");
     return kFALSE;
   }

  if ( fVarComboBox->GetNumberOfEntries() <= 0)
   {
     Error("Init","No parametric functions to show");
     return kFALSE;
   }


  
  fVarComboBox->Select(1);
  
  fDrawnFunctions = new TObjArray(fVarComboBox->GetNumberOfEntries(),0);
  
  fYaxisBuffer = new  Double_t [fNSPoints];
  fSaxis       = new  Double_t [fNSPoints];
  
  for (Int_t i=0; i< fNSPoints; i++)
   {
     Int_t ifort=i+1;
     fSaxis[i] = w_ptc_getlengthat(&ifort);
     Info("Init","Z position at %d is %f",i+1,fSaxis[i]);
   }

  Info("Init","Initialization Done");
  
  return kTRUE;

}
//______________________________________________________________________________

//______________________________________________________________________________

void MadxViewerPar::ConnectSignals2Slots()
{
   // Connect check buttons.
   if (fMaster == 0x0)
    {
      Error("ConnectSignals2Slots","Master Viewer is not set");
      return;
    }

   
   
   fDrawButton->Connect("Clicked()", "MadxViewerPar", this,"DrawParametric()");

//   fKnobOneMinEntry->Connect("ValueSet(Long_t)","MadxViewerPar", this,"SetParOneRange");
   
}

//______________________________________________________________________________

void MadxViewerPar::DrawParametric()
{
  
  if (fMaster == 0x0) return;
  
  fMaster->FocusOnCanvas();
  fMaster->SetActiveAlong();
  fMaster->HideKnobs(-1,-1);//show all knobs

  Int_t selectedVar  = fVarComboBox->GetSelected();
 
  Info("DrawParametric","Drawing %d variable",selectedVar);

  if (fSameButton->IsOn())
   {
     Info("DrawParametric","Drawing in the same window");
   }
  else
   {
     fLegend->Clear();
     fDrawnFunctions->Delete();
   }
  
  if (fDrawnFunctions->At(selectedVar)) 
   {
     Error("DrawParametric","This function is already drawn");
     return;
   }
   
  for (Int_t i = 1; i <= fNSPoints; i++)
   {
     fYaxisBuffer[i-1] = w_ptc_getfunctionvalueat(&selectedVar,&i);
     Info("DrawParametric","Function %d  at position %d is %f",selectedVar,i,fYaxisBuffer[i-1]);
   }
  
  TGraph* fctn = new TGraph(fNSPoints,fSaxis,fYaxisBuffer);
  fDrawnFunctions->AddAtAndExpand(fctn,selectedVar);
  
  fctn->SetEditable(kFALSE);
  fctn->SetLineWidth(3);
  fctn->SetFillStyle(0);
  fctn->SetFillColor(0);

  if ( fSameButton->IsOn() && (fDrawnFunctions->GetEntries() > 1) )
   {

     Int_t col = fDrawnFunctions->GetEntries();
     if (col > 10) col++;
     fctn->SetLineColor(col);
     fctn->Draw("Lsame");
     SetYRange();
     
   }
  else
   {
     fctn->Draw("AL");
     fYaxis = fctn->GetYaxis();
     fctn->GetXaxis()->SetTitle("s [m]");
   }
  
  
  fLegend->AddEntry(fctn,fVarComboBox->GetSelectedEntry()->GetTitle());
  fLegend->Draw();
  
  
  fMaster->UpdateCanvas();
  
}
//______________________________________________________________________________
void MadxViewerPar::UpdatePlots()
{
  for (Int_t j = 1; j < fDrawnFunctions->GetEntriesFast(); j++)
   {
     TGraph* g = dynamic_cast<TGraph*>(fDrawnFunctions->At(j));
     if (g == 0x0) continue;
     
     Double_t* y = g->GetY();
     for (Int_t i = 1; i <= fNSPoints; i++)
      {
        y[i-1] = w_ptc_getfunctionvalueat(&j,&i);
//        Info("UpdatePlots","Function %d  at position %d is %f",j,i,y[i-1]);
      }
   }
  SetYRange();
}
//______________________________________________________________________________

void MadxViewerPar::ReDraw()
{
  Info("ReDraw","");
  Int_t first = -1;
  for (Int_t j = 1; j < fDrawnFunctions->GetEntriesFast(); j++)
   {
     TGraph* g = dynamic_cast<TGraph*>(fDrawnFunctions->At(j));
     if (g == 0x0) continue;

     if ( fYaxis != g->GetYaxis() ) continue;
     
     Info("ReDraw","The first one");
     g->Draw("AL");
     first = j;
     break;
   }


  for (Int_t j = 1; j < fDrawnFunctions->GetEntriesFast(); j++)
   {
     TGraph* g = dynamic_cast<TGraph*>(fDrawnFunctions->At(j));
     if (g == 0x0) continue;
     if ( first == j ) continue;

     Info("ReDraw","%d",j);
     g->Draw("Lsame");
     
   }
   
  fLegend->Draw();
  fMaster->HideKnobs(-1,-1);//show all knobs
  fMaster->UpdateCanvas();  
}
//______________________________________________________________________________

void MadxViewerPar::Clear()
{
  Info("Clear","Cleaning");
  fLegend->Clear();
  fDrawnFunctions->Delete();

}
//______________________________________________________________________________


void MadxViewerPar::SetYRange()
 {
//Sets Y range in 1D functions

  if (fDrawnFunctions->GetEntries() == 0) return; //nothing to do
  if (fYaxis == 0x0) return; //nothing to do
  
  
  Double_t min = FLT_MAX;
  Double_t max =-FLT_MAX;


  TIter next(fDrawnFunctions);

  TGraph* gg = 0x0;

  while(( gg = (TGraph*)next() ))
   {
     Double_t* y = gg->GetY();
     if (y == 0x0)
      {
        Error("SetYRange","Y Array is NULL");
        continue;
      }
      
     for (Int_t i = 0; i < fNSPoints; i++)
      {
        if (y[i] > max) 
         {
           max = y[i];
         }

        if (y[i] < min) 
         {
           min = y[i];
         }
       }   
   }

//  Info("DrawParametricAt","Global: %f %f ",min,max);

  Double_t delta = TMath::Max(TMath::Abs(min),TMath::Abs(max));

  min = min - delta*0.2;
  max = max + delta*0.2;

//  Info("DrawParametricAt","New Limit: %f %f, delta %f",min,max,delta);
  fYaxis->SetRangeUser(min,max);
}
//______________________________________________________________________________

void MadxViewerPar::SetKnobs()
{
   
}
//______________________________________________________________________________


void MadxViewerPar::AddFunction(Int_t n, const char* name )
{
   if (fVarComboBox) fVarComboBox->AddEntry(name,n);
}



